# 
# -------------------------------------------------------------------
#                                   X-BONE
#
#                          http://www.isi.edu/xbone
#                USC Information Sciences Institute (USC/ISI)
#                   Marina del Rey, California 90292, USA
#                          Copyright (c) 1998-2005
# 
# -------------------------------------------------------------------
#
# Copyright (c) 1998-2005 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and
# its documentation in source and binary forms for non-commercial
# purposes and without fee is hereby granted, provided that the above
# copyright notice appear in all copies and that both the copyright
# notice and this permission notice appear in supporting
# documentation, and that any documentation, advertising materials,
# and other materials related to such distribution and use acknowledge
# that the software was developed by the University of Southern
# California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived
# from this software without specific prior written permission.
# 
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT
# THE SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS
# PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so
# noted when applicable.
#
# -------------------------------------------------------------------
#
# Effort partly sponsored by the Defense Advanced Research Projects
# Agency (DARPA) and Air Force Research Laboratory, Air Force Materiel
# Command, USAF, under agreement numbers F30602-98-1-0200 (X-Bone) and
# F30602-01-2-0529 (DynaBone). The views and conclusions contained
# herein are those of the authors and should not be interpreted as
# necessarily representing the official policies or endorsements,
# either expressed or implied, of the Defense Advanced Research
# Projects Agency (DARPA), the Air Force Research Laboratory, or the
# U.S. Government.
#
# This work was partly supported by the NSF STI-XTEND (ANI-0230789)
# and NETFS (ANI-0129689) projects. Any opinions, findings, and
# conclusions or recommendations expressed in this material are those
# of the authors and do not necessarily reflect the views of the
# National Science Foundation.
#
# ----------------------------------------------------------------------------
#  $RCSfile: HTTP.pm,v $
#
# $Revision: 1.4 $
#   $Author: pingali $
#     $Date: 2005/03/31 07:03:43 $
#    $State: Exp $
# ----------------------------------------------------------------------------

# Primary Author: Lars Eggert

package HTTP;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw(add handle $port $httpd);

use strict;
use Utilities qw(:DEFAULT $verbose $debug);
use Mux qw(%attacked_innerlays);

use LWP::MediaTypes qw(:DEFAULT add_type);
use HTTP::Daemon;
use HTTP::Status;
use HTTP::Response;
use Net::hostent;
use NetPacket::IP qw(:protos);

# Globals
#
our $port = 50159;
our $root_path = "$FindBin::Bin/webvis/";
our $httpd;

my %pkt_history;       # separate history for each web client
my %pkt_history_timer; # time of last data request from client
my $default = "default";


# Purpose:
#       Map an innerlay ID into an integer. For display purposes ONLY.
#       This will fail horribly if the BARP table changes during runtime.
#
#       Must use sysdie/syswarn; this runs after we became a daemon.
#
# Arguments:
#       $id    innerlay ID
#
# Returns:
#       Integer greater or equal to zero. 
#
# Exceptions:
#       thrown on error
#
sub innerlay_number ($) {
  my ($id) = @_;
  my @innerlays = BARP::innerlay_ids;
  for (my $i = 0; $i < scalar @innerlays; $i++) {
    return $i if $innerlays[$i] eq $id;
  }
  sysdie "could not find innerlay number for $id";
}


# Purpose:
#       Construct a textual message describing the packet history and
#       attack status since the last client query.
#
#       Must use sysdie/syswarn; this runs after we became a daemon.
#
# Arguments:
#       $peer   unique peer ID
#
# Returns:
#       Text string in the following format (in "perl-ized" EBNF):
#
#           MSG         := ATTACK ':' PACKET_INFO
#           ATTACK      := '-' | INNERLAY (',' INNERLAY)*
#           PACKET_INFO := '-' | INNERLAYS '@' DIRECTION '@' LENGTH '@' COLOR
#           INNERLAYS   := INNERLAY ('|' INNERLAY)*
#           DIRECTION   := 'r' | 's'
#           LENGTH      := [0-9]+
#           COLOR       := [a-z]+
#           INNERLAY    := [0-9]+
#
#       Examples: 
#           14,11:0|2@s@84@black,1@s@84@white,0@r@184@red,2|1|0@s@40@green
#           -:2|1|0@s@40@green
#           11:-
#           -:-
#
# Exceptions:
#       -
#
sub data_msg ($) {
  my ($peer) = @_;

  # if this is the first time we talk to this peer, initialize its
  # packet history from the default
  unless (exists $pkt_history{$peer}) {
    $pkt_history{$peer} = [ @{ $pkt_history{$default} } ];
    sysprint "New peer $peer" if $verbose;
  }
  
  # build attack info
  my $attack = (join "|", map { innerlay_number $_ } keys %attacked_innerlays);
  $attack = "-" unless length $attack;

  # build packet info
  my $packet_info = (join ",", @{ $pkt_history{$peer} });
  $packet_info = "-" unless length $packet_info;

  return "$attack:$packet_info";
}


# Purpose:
#       Open HTTP server listen socket $httpd.
#
# Arguments:
#       -
#
# Returns:
#       -
#
# Exceptions:
#       thrown on error
#
sub open () {
  # First, add some MIME types, so our files get send correctly.
  add_type("text/css" => qw(css));
  add_type("image/svg+xml" => qw(svg));
  # This would be RFC-compliant, but IE doesn't like it:
  # add_type("application/xhtml+xml" => qw(html));

  # Now open the socket.
  $httpd = new HTTP::Daemon LocalPort => $port, ReuseAddr => 1 or die $!;
  sysprint $httpd->url if $verbose; 

  # Init default packet history.
  $pkt_history{$default} = [];
  $pkt_history_timer{$default} = time;
}


# Purpose:
#       Send string $msg as HTTP response to client $c.
#
#       Must use sysdie/syswarn; this runs after we became a daemon.
#
# Arguments:
#       $c    client (HTTP::Request object)
#
#       $msg  string
#
# Returns:
#       -
#
# Exceptions:
#       -
#
sub respond ($$) {
  my ($c, $msg) = @_;
  my $res = new HTTP::Response;

  # set content and OK error code
  $res->content($msg);
  $res->code(RC_OK);

  # make sure this is never cached
  $res->expires(0);
  $res->header("Cache-Control" => "no-cache");
  $res->header("Pragmas" => "no-cache");

  # send it
  $c->send_response($res);
}


# Purpose:
#       Accept connection on the HTTP listen socket, and handle the
#       request, sending an appropriate response to the client.
#
#       Must use sysdie/syswarn; this runs after we became a daemon.
#
# Arguments:
#       -
#
# Returns:
#       -
#
# Exceptions:
#       thrown on error
#
sub handle () {
  # see http://use.perl.org/~jjohn/journal/9030
  local $SIG{PIPE} = "IGNORE";

  # accept and read request
  my $c = $httpd->accept;
  my $r = $c->get_request;

  if ($verbose) {
    my $h = gethost $c->peerhost;
    $h = ($h ? $h->name : $c->peerhost);
    sysprint $h .":". $c->peerport ." ". $r->method ." ". $r->uri;
  }
  
  # if it is a GET...
  if ($r->method eq "GET") {

    # and it is a GET for the packet history...
    if ($r->url->path eq "/data") {
      # extract query arguments
      my ($url, $param) = split "\\?", $r->uri;
      my ($tag, $val) = split "=", $param;

      # send the packet history response
      respond($c, data_msg $val);

      # now clear the packet history
      $pkt_history{$val} = [];

      # remember access time
      $pkt_history_timer{$val} = time;

    } elsif ($r->url->path eq "/policy") {
      # return the current policy
      respond($c, $Mux::policy);

    } else {
      # if it isn't a packet history request, try to see if it is for a file
      if ($r->url->path eq "/") { $r->url->path("/index.html"); }
      my $file = $root_path . substr $r->url->path, 1;
      if (-f $file) {
	# SECURITY ISSUE: any file under $root_path can be requested
	$c->send_file_response($file);
      } else {
	# couldn't find the file, send error code
	$c->send_error(RC_FORBIDDEN);
      }
    }
    
  # else, if this a POST...  
  } elsif ($r->method eq "POST") {
    # extract query arguments
    my ($url, $param) = split "\\?", $r->uri;
    my ($tag, $val) = split "=", $param;
    
    # see what URL the POST is going to...
    my @ids = BARP::innerlay_ids;
    if ($url eq "/stop") {
      # attack indication
      unless ($val < scalar @ids) {
	syswarn "Illegal innerlay number $val" if $verbose;
      } else {
	$attacked_innerlays{$ids[$val]} = 1;
	$c->send_response;
	sysprint "Innerlay number $val (ID $ids[$val]) blocked" if $verbose;
      }
      
    } elsif ($url eq "/go") {
      # clear attack status
      unless ($val < scalar @ids) {
	syswarn "Illegal innerlay number $val" if $verbose;
      } else {
	delete $attacked_innerlays{$ids[$val]};
	$c->send_response;
	sysprint "Innerlay number $val (ID $ids[$val]) unblocked" if $verbose;
	}
      
    } elsif ($url eq "/policy") {
      # change policy (XXX check validity!)
      $Mux::policy = $val;
      $c->send_response;
      sysprint "Mux'ing policy set to $Mux::policy" if $verbose;
      
    } else {
      # no idea what this is, send error code
      $c->send_error(RC_FORBIDDEN);
    }

  # else it's neither a POST, nor a GET, so send an error response
  } else {
    $c->send_error(RC_FORBIDDEN);
  }
  
 done:
  # close the connection - we don't do pipelining, since the accept
  # could block, and then we'd not handle outerlay traffic, since our
  # select loop is somewhat simple
  $c->close;
}


# Purpose:
#       Add a packet to the packet history.
#
#       Must use sysdie/syswarn; this runs after we became a daemon.
#
# Arguments:
#       $length      packet length in bytes
#
#       $type        packet types are magic (need to know which will come in)
#
#       $direction   "s" for send, "r" for receive
#
#       @innerlays   list of innerlay IDs the packet was sent/received on
#
# Returns:
#       -
#
# Exceptions:
#       thrown on error
#
sub add ($$$@) {
  my ($length, $type, $direction, @innerlays) = @_;
  my %colors = ( IP_PROTO_UDP()  => "cyan",    IP_PROTO_TCP()  => "green",
		 IP_PROTO_ICMP() => "magenta", IP_PROTO_IGMP() => "orange",
		 IP_PROTO_IP()   => "tan",     IP_PROTO_IPIP() => "yellow");

  # sanity checks
  sysdie "Illegal direction $direction" unless $direction =~ /^[sr]$/;
  sysdie "No innerlays" unless scalar @innerlays > 0;
  sysdie "Illegal length $length" unless $length > 0;

  # find innerlay number
  my @innerlay_numbers = map { innerlay_number $_ } @innerlays;

  # set color (make black if we don't have enough defined ones)
  my $color = "black";
  $color = $colors{$type} if exists $colors{$type};

  # update all existing histories
  foreach my $peer (keys %pkt_history) {
    # remove peers (not default) that didn't retrieve their history in a while
    if ($peer ne $default and $pkt_history_timer{$peer} + 300 < time) {
      sysprint "Peer $peer timed out" if $verbose;
      delete $pkt_history_timer{$peer};
      delete $pkt_history{$peer};
      next;
    }

    # never keep more than 100 packets in the history
    shift @{ $pkt_history{$peer} } while scalar @{ $pkt_history{$peer} } > 100;
    
    # add to history
    push @{ $pkt_history{$peer} }, 
      (join "@", ((join "|", @innerlay_numbers), $direction, $length, $color));
  }
}


1;
