# 
# -------------------------------------------------------------------
#                                   X-BONE
#
#                          http://www.isi.edu/xbone
#                USC Information Sciences Institute (USC/ISI)
#                   Marina del Rey, California 90292, USA
#                          Copyright (c) 1998-2005
# 
# -------------------------------------------------------------------
#
# Copyright (c) 1998-2005 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and
# its documentation in source and binary forms for non-commercial
# purposes and without fee is hereby granted, provided that the above
# copyright notice appear in all copies and that both the copyright
# notice and this permission notice appear in supporting
# documentation, and that any documentation, advertising materials,
# and other materials related to such distribution and use acknowledge
# that the software was developed by the University of Southern
# California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived
# from this software without specific prior written permission.
# 
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT
# THE SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS
# PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so
# noted when applicable.
#
# -------------------------------------------------------------------
#
# Effort partly sponsored by the Defense Advanced Research Projects
# Agency (DARPA) and Air Force Research Laboratory, Air Force Materiel
# Command, USAF, under agreement numbers F30602-98-1-0200 (X-Bone) and
# F30602-01-2-0529 (DynaBone). The views and conclusions contained
# herein are those of the authors and should not be interpreted as
# necessarily representing the official policies or endorsements,
# either expressed or implied, of the Defense Advanced Research
# Projects Agency (DARPA), the Air Force Research Laboratory, or the
# U.S. Government.
#
# This work was partly supported by the NSF STI-XTEND (ANI-0230789)
# and NETFS (ANI-0129689) projects. Any opinions, findings, and
# conclusions or recommendations expressed in this material are those
# of the authors and do not necessarily reflect the views of the
# National Science Foundation.
#
# -------------------------------------------------------------------
#  $RCSfile: IPIP.pm,v $
#
# $Revision: 1.4 $
#   $Author: pingali $
#     $Date: 2005/03/31 07:03:44 $
#    $State: Exp $
# ----------------------------------------------------------------------------

# Primary Author: Lars Eggert

package IPIP;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw(inbound outbound $ttl);

use strict;
use Utilities qw(:DEFAULT $verbose $debug);
use Mux qw(%attacked_innerlays);
use BARP;
use NetPacket::IP qw(:DEFAULT :protos :flags :versions IP_MAXPACKET);
use NetPacket::TCP;
use NetPacket::UDP;
use NetPacket::ICMP;
use Net::protoent;
use Net::servent;


# Globals
#
my $ttl       = sysctl_read "net.inet.ip.ttl";   # TTL for tunnel wrapper


# Purpose:
#       Return a string that describes the argument packet for debugging.
#
# Arguments:
#       $pkt    packet to describe
#
# Returns:
#       Textual description of the packet.
#
# Exceptions:
#       thrown on error
#
sub describe_pkt ($) {
  
  sub translate_port ($$) {
    my ($port, $proto) = @_;
    return "" unless defined $port;
    my $serv = getservbyport ($port, $proto);
    if (defined $serv) {
      return sprintf ":%d (%s)", $port, $serv->name;
    }
    return sprintf ":%d", $port;
  } 
 
  my ($pkt) = @_;

  # do nothing if not verbose
  return "" unless $verbose;
  
  my ($sport, $dport, $tpkt, $flag_desc);
  
  if ($pkt->{proto} == IP_PROTO_UDP) {
    $tpkt = NetPacket::UDP->decode($pkt->{data});

  } elsif ($pkt->{proto} == IP_PROTO_TCP) {
    $tpkt = NetPacket::TCP->decode($pkt->{data});
    my %flags_desc = (@{[FIN]} => "FIN", @{[SYN]} => "SYN",
		      @{[RST]} => "RST", @{[PSH]} => "PSH", 
		      @{[ACK]} => "ACK", @{[URG]} => "URG",
		      @{[ECE]} => "ECE", @{[CWR]} => "CWR");
    my @flags = ();
    foreach my $f (FIN, SYN, RST, PSH, ACK, URG, ECE, CWR) {
      if ($tpkt->{flags} & $f) { push @flags, $flags_desc{$f}; }
    }
    $flag_desc = "<" . (join "|", @flags) . ">";
  }

  $sport = $tpkt->{src_port};
  $dport = $tpkt->{dest_port};
  $flag_desc = "" unless defined $flag_desc;
  
  my $proto = getprotobynumber $pkt->{proto};
  $proto = $proto->name;
  
  return sprintf "%s%s id %d%s%s %s%s -> %s%s", uc $proto, $flag_desc,
    $pkt->{id}, ($pkt->{flags} & IP_FLAG_DONTFRAG ? " DF" : ""),
    ($pkt->{len} != -1 ? " len ". $pkt->{len} : ""),
    $pkt->{src_ip}, translate_port($sport, $proto), 
    $pkt->{dest_ip}, translate_port($dport, $proto);
}


# Purpose:
#       Outbound processing - read outgoing outerlay packet from tun
#       device, pick an innerlay to multiplex over, encapsulate (=
#       create an innerlay paacket) and re-inject the result into the
#       kernel.
#
#       Must use sysdie/syswarn; this runs after we became a daemon.
#
# Arguments:
#       $fh             file handle to the tun device from which to read the
#                       outgoing packet
#
#       $outerlay_block IP block (Net::IP object) of the outerlay interface
#
#       $outerlay_addr  outerlay IP address of this host (Net::IP object)
#
# Returns:
#       -
#
# Exceptions:
#       thrown on error
#
sub outbound ($$$) {
  my ($tun, $outerlay_block, $outerlay_addr) = @_;
  my $raw_pkt = "";

  # Read the outbound packet from the tun device (file handle must be
  # non-blocking, otherwise the read call waits until the passed
  # buffer is full, instead of returning exactly one packet - XXX
  # investigate whether recv works instead of read here)
  sysread $tun, $raw_pkt, IP_MAXPACKET 
    or sysdie "Cannot read packet from tun: $!";
  my $pkt = NetPacket::IP->decode($raw_pkt);

  # Sanity check: RFC 2003 says MUST NOT encapsulate when TTL=0
  if ($pkt->{ttl} == 0) {
    syswarn "Must not encapsulate pkt with ttl=0" if $verbose;
    return;
  }

  # Drop packets destined to the subnet broadcast address of the
  # outerlay.
  if ($pkt->{dest_ip} eq $outerlay_block->last_ip) {
    syswarn "Dropping broadcast: ". describe_pkt $pkt if $verbose;
    return;
  }

  # Drop packets that aren't IPv4.
  if ($pkt->{ver} != IP_VERSION_IPv4) {
    syswarn "Dropping non-IPv4 packet (IP version $pkt->{ver})" if $verbose;
    return;
  }

  # Pick innerlay(s) to forward this packet over.
  my @selected_innerlays = Mux::mux $pkt;

  # If muxing returns an empty list of innerlays to forward, drop it.
  if (scalar @selected_innerlays == 0) {
    if ($verbose) {
      my $pdesc = describe_pkt $pkt;
      sysprint "Drop $pdesc outbound innerlay set empty.";
    }
    return;
  }

  # We know which innerlays to encapsulate the packet over, do so now. 
  #
  # THIS CODE MUST DO RFC 2003-CONFORMING IP-IN-IP ENCAPSULATION! (And
  # tries to, but may not cover all borderline cases. Anything
  # different from RFC 2003 should be considered a bug.)
  #
  # The biggest issue we dodge is PMTU discovery-related things, like
  # ICMP relaying/generation. (This is mostly because it's hard to do
  # in userland.) Settig DF on the outer header in all packets seems to
  # help with some PMTU blackholing, but this is still a *major*
  # issue that needs re-thinking.
  #
  # We're reusing a copy of the headers of the original packet, and
  # diddle the fields that are different.
  my $encap_pkt = NetPacket::IP->decode($raw_pkt);
  $encap_pkt->{proto}   = IP_PROTO_IPIP;
  $encap_pkt->{data}    = $raw_pkt;
  $encap_pkt->{flags}   |= IP_FLAG_DONTFRAG;  # see section 5.1 of RFC 2003
  $encap_pkt->{options} = "";   # RFC 2003 says don't copy IP options
  $encap_pkt->{ttl}     = $ttl; # use specified TTL
  $encap_pkt->{id}      = int rand 65536;
  $encap_pkt->{len}     = -1;   # set by NetPacket::IP, fake so I remember

  # forward over all selected innerlays.
  my %local_ips = BARP::innerlay_ips $outerlay_addr;
  my %remote_ips = BARP::innerlay_ips new Net::IP $pkt->{dest_ip};
  foreach my $innerlay (@selected_innerlays) {
    # use correct innerlay source and destination for each duplicate
    $encap_pkt->{src_ip}  = $local_ips{$innerlay}->ip;
    $encap_pkt->{dest_ip} = $remote_ips{$innerlay}->ip;

    # constructed the encapsulated packet, insert it into the
    # kernel (NetPacket::IP::encode automatically computes the checksum
    # and sets the packet length)
    syswrite $tun, $encap_pkt->encode;
  }
  
  # debug info
  if ($verbose) {
    my $pdesc = describe_pkt $pkt;
    my $edesc = describe_pkt $encap_pkt;
    syswarn "Outbound $pdesc inside $edesc over " .
      (join ", ", @selected_innerlays);
  }
  
  # maintain packet history
  HTTP::add($pkt->{len}, $pkt->{proto}, "s", @selected_innerlays);
}



# Purpose:
#       Inbound processing - read an incoming innerlay packet from the
#       divert socket, validate the encapsulation, decapsulate,
#       validate the contents (= outerlay packet) and re-inject the
#       result back into the kernel.
#
#       To filter out redundant duplicates arriving on different
#       innerlays, this function will not immediately re-inject a
#       packet on reception, but cache it until a new packet arrives
#       that isn't a duplicate. (This assumes redundant duplicates
#       arrive back-to-back over different innerlays.)
#
#       Must use sysdie/syswarn; this runs after we became a daemon.
#
# Arguments:
#       $fh          file handle of divert socket from which to read the
#                    incoming packet
#
#       $outerlay_addr  outerlay IP address of this host (Net::IP object)
#
# Returns:
#       -
#
# Exceptions:
#       thrown on error
#

my $last_ip_id = -1;  # These globals are "static" for inbound(), and
my @last_innerlays;   # maintain the last seen packet information.
my $last_pkt;         # (XXX why does "local" not work anymore with
my $last_decap_pkt;   # perl5.8? strange)

sub inbound ($$) {
  my ($divert, $outerlay_addr) = @_;
  my $raw_pkt = "";

  # a packet is arriving over the innerlay, via a divert rule - read it
  my $tag = recv $divert, $raw_pkt, IP_MAXPACKET, 0 
    or sysdie "Cannot read packet from divert: $!";
  my $pkt = NetPacket::IP->decode($raw_pkt);

  # XXX RFC 2003 may require additional checks here - add them!
  # XXX we should validate whether the incoming packets satisfy the
  # muxing policy in outbound() - we don't do this yet

  # sanity check: make sure pkt is for one of our local innerlay
  # addresses (pkt dest matches one of our local innerlay dests)
  my $arrived_on;
  my %innerlays = BARP::innerlay_ips $outerlay_addr;
  foreach my $id (BARP::innerlay_ids) {
    if ($innerlays{$id}->ip eq $pkt->{dest_ip}) { $arrived_on = $id; last; }
  }
  unless (defined $arrived_on) {
    syswarn "Got pkt for non-local innerlay dst $pkt->{dest_ip}" if $verbose;
    goto REINJECT;
  }

  # looks like the packet wrapper is OK, decapsulate it
  my $decap_pkt = NetPacket::IP->decode($pkt->{data});

  # sanity check: make sure pkt is for our outerlay address
  unless ($decap_pkt->{dest_ip} eq $outerlay_addr->ip) {
    syswarn "Got pkt for non-local outerlay dst $decap_pkt->{dest_ip}"
      if $verbose;
    goto REINJECT;
  }

  # sanity check: RFC 2003 says MUST NOT deliver pkt when TTL=0
  if ($decap_pkt->{ttl} == 0) {
    syswarn "Must not deliver pkt with ttl=0" if $verbose;
    return; # and not "goto REINJECT" - this is an invalid packet for us
  }

  # If the packet arrived over an attacked innerlay, we must drop it.
  if (exists $attacked_innerlays{$arrived_on}) {
    if ($verbose) {
      my $pdesc = describe_pkt $pkt;
      sysprint "Drop $pdesc inbound on attacked innerlay $arrived_on."
      }
    return; # and not "goto REINJECT" - this was for us, but we must drop
  }
  
  # since we now support redundant copies on parallel innerlays, make
  # sure to only pass back the first such copy, drop all others
  if ($decap_pkt->{id} == $last_ip_id) {
    # this is a duplicate of the last packet we received
    # keep track of list of innerlays we've seen this packet come in on
    push @last_innerlays, $arrived_on;
    return; # and not "goto REINJECT" - this was for us, but a duplicate

  } else {
    # this is a new packet, i.e. not a duplicate of the last one
    
    if ($last_ip_id != -1) {
      # if this is not the very first packet, maintain packet history
      # of the PREVIOUSLY seen packet
      HTTP::add($last_decap_pkt->{len}, $last_decap_pkt->{proto},
		"r", @last_innerlays);
      
      # debug info
      if ($verbose) {
	my $pdesc = describe_pkt $last_pkt;
	my $ddesc = describe_pkt $last_decap_pkt;
	syswarn "Inbound $ddesc inside $pdesc over " .
	  (join ", ", @last_innerlays);
      }
    }
    
    # This injects the packet as soon as it is received, i.e. the
    # first duplicate. Pro: No traffic delays wainting for the first
    # non-duplicate to come in. Contra: Packet history is distored,
    # because the first inbound copy could trigger a response before
    # the packet history is updated.
    #
    # (XXX don't use $divert->send syntax, it injects as outbound!)
    send $divert, $pkt->{data}, 0, $tag or sysdie "Cannot inject packet: $!";
    
    # this is no duplicate, so remember it
    @last_innerlays = ($arrived_on);
    $last_ip_id = $decap_pkt->{id};
    $last_decap_pkt = $decap_pkt;
    $last_pkt = $pkt;
    return; # and not "goto REINJECT" - was for us, but may be more duplicates
  }

 REINJECT:
  # Re-inject the original raw packet - it might be for someone else!
  send $divert, $raw_pkt, 0, $tag or sysdie "Cannot re-inject raw packet: $!";
}




1;
