# 
# -------------------------------------------------------------------
#                                   X-BONE
#
#                          http://www.isi.edu/xbone
#                USC Information Sciences Institute (USC/ISI)
#                   Marina del Rey, California 90292, USA
#                          Copyright (c) 1998-2005
# 
# -------------------------------------------------------------------
#
# Copyright (c) 1998-2005 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and
# its documentation in source and binary forms for non-commercial
# purposes and without fee is hereby granted, provided that the above
# copyright notice appear in all copies and that both the copyright
# notice and this permission notice appear in supporting
# documentation, and that any documentation, advertising materials,
# and other materials related to such distribution and use acknowledge
# that the software was developed by the University of Southern
# California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived
# from this software without specific prior written permission.
# 
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT
# THE SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS
# PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so
# noted when applicable.
#
# -------------------------------------------------------------------
#
# Effort partly sponsored by the Defense Advanced Research Projects
# Agency (DARPA) and Air Force Research Laboratory, Air Force Materiel
# Command, USAF, under agreement numbers F30602-98-1-0200 (X-Bone) and
# F30602-01-2-0529 (DynaBone). The views and conclusions contained
# herein are those of the authors and should not be interpreted as
# necessarily representing the official policies or endorsements,
# either expressed or implied, of the Defense Advanced Research
# Projects Agency (DARPA), the Air Force Research Laboratory, or the
# U.S. Government.
#
# This work was partly supported by the NSF STI-XTEND (ANI-0230789)
# and NETFS (ANI-0129689) projects. Any opinions, findings, and
# conclusions or recommendations expressed in this material are those
# of the authors and do not necessarily reflect the views of the
# National Science Foundation.
#
# ----------------------------------------------------------------------------
#  $RCSfile: Mux.pm,v $
#
# $Revision: 1.4 $
#   $Author: pingali $
#     $Date: 2005/03/31 07:03:44 $
#    $State: Exp $
# ----------------------------------------------------------------------------

# Primary Author: Lars Eggert

package Mux;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(mux);
@EXPORT_OK = qw(%attacked_innerlays $policy);

use strict;
use Utilities qw(:DEFAULT $verbose $debug);
use BARP;


# Globals - must be initialized before mux may be called
#
our %attacked_innerlays; # list of innerlay IDs currently under attack
our $policy = "seesaw";  # current mux'ing policy

my $round_robin_pos = 0;
my $seesaw_pos = 0;
my $seesaw_step = 1;


# Purpose:
#       Factor an integer into a list of its prime factors. Adapted from
#       http://www.mamiyami.com/docs/perl/cookbook/ch02_20.htm
#
# Arguments:
#       $n   integer number to factor
#
# Returns:
#       Return list of prime factors. Multiplying all elements together
#       yields $n.
#
# Exceptions:
#       thrown on error
#
sub prime_factors {
  my ($n) = @_;
  my @result = qw(1);

  return qw(1) if $n == 1;

  # count up to the first factor
  my $i = 2;
  $i++ while $n % $i;
  push @result, $i;

  # it's prime, no more factors
  return @result if $i == $n;

  # tail recurse
  push @result, prime_factors($n / $i);
  return @result;
}


# Purpose:
#       Pick innerlay(s) to multiplex over.
#
# Arguments:
#       $pkt            packet to multiplex
#
# Returns:
#       Return list of selected innerlay IDs.
#
# Exceptions:
#       thrown on error
#
sub mux ($) {
  my ($pkt) = @_;
  my @which; # is a list of innerlay NUMBERS, that are mapped to IDs on exit

  # Create a list of currently available innerlays, i.e. all that
  # aren't int he attack list.
  my @ok_innerlays = 
    grep { not exists $attacked_innerlays{$_} } BARP::innerlay_ids;

  if ($policy eq "random") {
    # This policy completely randomizes packets over the innerlays.
    push @which, int rand scalar @ok_innerlays;

  } elsif ($policy eq "ip-id") {
    # This policy spreads packets into innerlays modulo their IP ID.
    # (On systems where IP IDs are randomized, this is very similar to
    # "random".)
    push @which, $pkt->{id} % scalar @ok_innerlays;

  } elsif ($policy eq "roundrobin") {
    # This policy does a round-robin on available overlay IDs.
    push @which, $round_robin_pos++;
    $round_robin_pos = 0 if $round_robin_pos >= scalar @ok_innerlays;

  } elsif ($policy eq "seesaw") {
    # This policy climbs up and down the available innerlays. 
    push @which, $seesaw_pos;
    $seesaw_pos += $seesaw_step;
    if ($seesaw_pos <= 0 and $seesaw_step < 0) {
      $seesaw_step = 1; 
      $seesaw_pos = 0;
    } elsif ($seesaw_pos >= scalar @ok_innerlays and $seesaw_step > 0) {
      $seesaw_step = -1;
      $seesaw_pos = scalar @ok_innerlays - 2;
    }

  } elsif ($policy eq "factors") {
    # This policy sends duplicates of a packet on the innerlay numbers
    # that form are the prime factors of its IP ID.
    @which = map { $_ - 1 } prime_factors $pkt->{id};

    # remove duplicate prime factors
    my %seen;
    @which = grep { not $seen{$_}++ } @which;

  } else {
    # this should never happen
    sysdie "Illegal policy $policy";
  }

  # remove picked innerlays outside the available range, and return IDs
  @which = grep { $_ >= 0 and $_ < scalar @ok_innerlays } @which;
  return map { $ok_innerlays[$_] } @which;
}


1;
